//**************************************************************************
//
//           T E C H N I C K A   U N I V E R Z I T A   K O S I C E
//
//                     Katedra pocitacov a informatiky
//
//
// AUTOR: Ladislav Vastag
// ODBOR: Vypoctova technika a informatika - Informacne technologie a web
// DATUM: 07.04.2004
// PROGRAM: Interpreter jednoduchecho jazyka podobneho BASIC-u 
//          z kapitoly 1.5.2 knihy "Navrh a interpretacia programovacich
//          jazykov", ktoru napisal doc. Kollar
//
// SUBORY: basint.c - (tento subor) obsahuje hlavnu funkciu
//                  - lexikalny analyzator realizovany funkciou getSymbol
//                  - syntakticky analyzator a interpreter, prepis EBNF
//                    do funkcii a pridanie denotacnej semantiky pre jed-
//                    notlive lexikalne jednotky (funkcie: Operand(), Mul(),
//                    Expr(), Stat(), StatSeq() a Program())
//
// POUZITA LITERATURA:
// - doc. Ing. Jan Kollar CSc. : Navrh a interpretacia programovacich jazykov
// - Ing. Pavel Herout : Ucebnica jazyka C
//
//**************************************************************************

#include <stdio.h>
#include <string.h>


#define   MAXIDLEN    8   // Max. dlzka identifikatora
#define   MAXIDS    100   // Max. pocet identifikatorov
#define   MAXNUMS   100   // Max. pocet konstant

#define   TRUE      1
#define   FALSE     0


// definovanie typu tabulky identifikatorov
typedef struct {
   int n;
   char tab[MAXIDS][MAXIDLEN];
}IDTABLE;


// definovanie typu tabulky konstant
typedef struct {
   int n;
   int tab[MAXNUMS];
}CONSTTABLE;


// mozina terminalnych symbolov definovanych cez enumeracny typ
typedef enum {
                SID,            // identifikator
                SCONST,         // konstanta
                LPAIR,          // '('
                RPAIR,          // ')'
                SASSIGN,        // '='
                SADD,           // '+'
                SSUB,           // '-'
                SMUL,           // '*'
                SDIV,           // '/'
                SSEMICOLON,     // ';'
                SPERIOD,        // '.'
                SEOF,           // '$'
                SINPUT,        // klucove slovo "input"
                SPRINT         // klucove slovo "print"
} SYMBOLS;


// definovanie typu klucoveho slova - je charakterizovane hodnotou  a typom
typedef struct {
   char kwstring[MAXIDLEN];    // hodnota klucoveho slova
   SYMBOLS symvalue;           // typ symbolu
}KEYWORD;

int Operand(void);
int Mul(void);
int Expr(void);
void Stat(void);
void StatSeq(void);
void Program(void);
char getCHARF(void);
void getSymbol(void);

// Deklaracia globalnych premennych
// Globalne premenne su pouzite pre vacsiu prehladnost programu a aby sa co
// najviac priblizil prikladu z knihy doc. Kollara, ktory je napisany
// v Pascale.
// Kto ma vrodeny odpor voci globalnym premennym moze sa pokusit program 
// prepisat. Staci "zabalit" vsetky globalne premenne do stuktury a potom vsetky
// funkcie volat s jednym parametrom a to smernikom na danu strukturu. Ovsem
// je nutne upravit aj samotne funkcie :-)


KEYWORD kwtab[2];             //Vytvorenie tabulky klucovych slov
IDTABLE tid;                   //Vytvorenie tabulky identifikatorov
CONSTTABLE ctab;               //Vytvorenie tabulky konstant
int mem[MAXIDS];               //Tabulka hodnot identifikatorov
FILE *p_srcfile;               //Smernik na zdrojovy subor
SYMBOLS symbol;                //nacitany symbol a pre SID a SCONST aj
int attr;                      //jeho atribut
int expr_res;                  //medzivysledok yatvorkovenho vyrazu

//==============================================================================
//                               main
//==============================================================================
main(int argc, char *argv[])
{
	int i;


	if(argc != 2)
	  {
	     printf("\n\n\tNESPRAVNE SPUSTENIE INTERPRETERA!!!");
	     return(1);
	  }

	// Inicializacia tabulky hodnot identifikatorov
	for(i=0;i<MAXIDS;i++)
	   mem[i]=0;
	p_srcfile=fopen(argv[1],"r");

	// Inicializacia
	tid.n=0;
	ctab.n=0;
	strcpy(kwtab[0].kwstring,"input");
	kwtab[0].symvalue=SINPUT;
	strcpy(kwtab[1].kwstring,"print");
	kwtab[1].symvalue=SPRINT;

	getSymbol();
	Program();
	fclose(p_srcfile);
	printf("\n\n");
	while(getchar()=='\r');
	getchar();
	return (0);
}

//==============================================================================
//                      SYNTAXOU RIADENY INTERPRETATOR
//
// Prepisane a OPRAVENE funkcie z kapitoly 3.2 Syntaxou riadeny interpretator
// knihy od doc. Kollara.
//==============================================================================

//================================= Operand ====================================
int Operand(void)
{
	int k;

   switch(symbol)
      {
	 case SID:    k=(mem[attr]);
		      getSymbol();
		      break;
	 case SCONST: k=(ctab.tab[attr]);
		      getSymbol();
		      break;
	 case LPAIR:  getSymbol();
		      k=Expr();
		      break;
	 case RPAIR:  k=expr_res;
		      getSymbol();
      }
   return k;
}

//=============================== Mul ==========================================
int Mul(void)
{
   int v1, v2;
	 SYMBOLS sy;

   v1=Operand();
   while(symbol == SMUL || symbol ==SDIV)
      {
         sy=symbol;
         getSymbol();
         v2=Operand();
         switch(sy)
            {
                case SMUL: v1=v1*v2;
                           break;
                case SDIV: v1=v1/v2;
            }
      }
   return v1;
}

//=============================== Epr ==========================================
int Expr()
{
	int v1, v2;
	SYMBOLS sy;

   v1=Mul();
   while(symbol==SADD || symbol==SSUB)
      {
	 sy=symbol;
	 getSymbol();
	 v2=Mul();
	 switch(sy)
	    {
	       case SADD: v1=v1+v2;
			  break;
	       case SSUB: v1=v1-v2;
	    }
      }

   // Koniec zatvoriek => predpokladame dalsi vyraz za za zatvorkou
   // volame teda Expr() a pocitame a pocitame ..... kym na nepretecu
   // nervy alebo zasobnik ;-)
   // Tu vsak este musim dodat jednu dolezitu vec, teda vlastne dve:
   //
   // 1.) Ak sa nasledujuci if vynecha, tak ako je to v knihe, program vypocita 
   // zatvorkovany vyraz len po prvu pravu zatvorku a zvysok vyrazu, ba co viac 
   // aj zvysok interpretovaneho zdrojoveho kodu jednoducho ignoruje.
   //
   // 2.) Po zaradeni nasl. if-u sa sice vyraz dopocita az do konca, ale 
   // priorita asociativnosti operatorov nemusi byt dodrzana. Teda vysledok
   // bude NESPRAVNY!!!
   if(symbol == RPAIR)
      {
	 expr_res=v1;
	 v1=Expr();
      }
   return v1;
}

//================================== Stat ======================================
void Stat()
{
		  int adr;

   switch(symbol)
      {
         case SINPUT: printf("INPUT ");
                      getSymbol();
                      printf("%s[%d] : ",tid.tab[attr],attr);
                      scanf("%d",&(mem[attr]));
		      getSymbol();
		      break;
	 case SPRINT: printf("PRINT ");
		      getSymbol();
		      printf("%d\n",Expr());
		      break;
         case SID:    adr=attr;
                      getSymbol();      //Dalsi citany symbol by mal byt '=', tu asi bude nutne doplnit kontrolu syntaxe
		      getSymbol();      //Mal by nasledovat vyraz.
		      mem[adr]=Expr();  //Vypocita sa vyraz
		      printf("%s[%d] = %d\n",tid.tab[adr],adr,mem[adr]);
                      break;
      }
}

//============================= StatSeq ===================================
void StatSeq()
{
   Stat();
   while(symbol==SSEMICOLON)
      {
         getSymbol();
         Stat();
      }
}

//================================= Program ====================================
void Program()
{
   StatSeq();
   // V tomto bode by mal byt symbol == SPERIOD
   getSymbol();
   // Vtomto bode by mal byt symbol == SEOF
   printf("STOP\n");
}


//==============================================================================
//                          LEXIKALNY ANALYZATOR
//
//==============================================================================

//=============================== getSymbol ====================================
void getSymbol(void)
{
	static int geton=TRUE;
	int lexoff, cval, k;
	char id[MAXIDLEN];
	static char ch;

  do{
       if(geton == TRUE)
          ch=getCHARF();
       geton=TRUE;
       lexoff=FALSE;

       switch(ch)
          {
              // SPECIALNE SYMBOLY
             case '=': symbol=SASSIGN;
                        break;
             case ';': symbol=SSEMICOLON;
                        break;
             case '.': symbol=SPERIOD;
                       break;
             case '$': symbol=SEOF;
                       break;
             case '(': symbol=LPAIR;
                       break;
             case ')': symbol=RPAIR;
                       break;
             case '+': symbol=SADD;
                       break;
             case '-': symbol=SSUB;
                       break;
             case '*': symbol=SMUL;
                       break;
             case '/': symbol=SDIV;
                       break;
             // KONSTANTY
             case '0':
             case '1':
             case '2':
             case '3':
             case '4':
             case '5':
             case '6':
             case '7':
             case '8':
             case '9': cval=ch-'0';
                       ch=getCHARF();
                       while('0'<=ch && ch<='9')
                          {
                             // Cita znaky konstanty a zaroven vypocitava jej hodnotu
                             cval=cval*10+ch-'0';
                             ch=getCHARF();
                          }
                       geton=FALSE;
                       symbol=SCONST;
                       k=0;
		       while((k<ctab.n)&&(cval!=ctab.tab[k]))
                          k++;
                       attr=k;            // k ukazuje na polozku v ctab.tab
                       if(k==ctab.n)
                         {  
                            // Nacitana konstanta este nie je v ctab.tab => ulozime
                            ctab.tab[k]=cval;
                            ctab.n=k+1;
                         }
                       break;
             // IDENTIFIKATORY
             case 'A':
             case 'B':
             case 'C':
             case 'D':
             case 'E':
             case 'F':
             case 'G':
             case 'H':
             case 'I':
             case 'J':
             case 'K':
             case 'L':
             case 'M':
             case 'N':
             case 'O':
             case 'P':
             case 'Q':
             case 'R':
             case 'S':
             case 'T':
             case 'U':
             case 'V':
             case 'W':
             case 'X':
             case 'Y':
             case 'Z':
             case 'a':
             case 'b':
             case 'c':
             case 'd':
             case 'e':
             case 'f':
             case 'g':
             case 'h':
             case 'i':
             case 'j':
             case 'k':
             case 'l':
             case 'm':
             case 'n':
             case 'o':
             case 'p':
             case 'q':
             case 'r':
             case 's':
             case 't':
             case 'u':
             case 'v':
             case 'w':
             case 'x':
             case 'y':
             case 'z': id[0]=ch;
                       k=0;
                       ch=getCHARF();
                       while(('0'<=ch && ch<='9') || ('A'<=ch && ch<='Z') || ('a'<=ch && ch<='z'))
                          {
                             k++;
                             id[k]=ch;
                             ch=getCHARF();
                          }
                       id[k+1]='\0';
                       geton=FALSE;
                       k=0;
                       
                       // Prezrieme tabulku klucovych slov ci nacitany symbol nie klucovym slovom
                       while((k<2) && (strcmp(id,kwtab[k].kwstring) !=0))
                          k++;
                       if(k<2)
                          symbol=kwtab[k].symvalue;      // v id je klucove slovo
                       else
                          {  
                             // Sem ideme ak v id nie je klucove slovo
                             k=0;
                             symbol=SID;
                             
                             // Hladame identifikator v tabulke
                             while((k<tid.n) && (strcmp(id,tid.tab[k])!=0))
                                k++;
                             attr=k;        // k ukazuje na polozku v tid.table
                             if(tid.n==k)
                                {  
                                   // Nacitany identifikator nie je este v tabulke => ulozime
                                   strcpy(tid.tab[k],id);
                                   tid.n=k+1;
                                }
                          }
                       break;
             
	     // KOMENTAR
	     case '\'': do{
			     ch=getCHARF();    // Citame komentra kym nenarazime na znak '
			     // POZOR!!! Nalsedujuci if v knihe nie je!!!!. Taktiez
                             // tam nie je kontrola dosiahnutia EOF v podmienke while.
                             // Dosledok ???
			     // Ak niektory komentar nie je uzavrety znakom ' , program
			     // sa zacykli!!!!!!
			     if(ch=='$')
				geton=FALSE;
			  }while(ch!='\'' && ch!='$');
			lexoff=TRUE;
			break;
             // MEDZERA
             case ' ': lexoff=TRUE;
                       break;
             // NEPRIPUSTNY ZNAK
             default:  break;            // Tu mi cosi chyba!!!!!!!!!!!!!
          }
    }while(lexoff);
}

//================================= getCHARF ===================================
char getCHARF(void)
{
	char c;

	c=getc(p_srcfile);
	if(feof(p_srcfile) != 0)
	  return '$';
	if(c == '\n')
	  return ' ';
	else
	  return c;
}
// ***  KONIEC  ***  Uff :-)